import strawberry
from django.contrib.auth import authenticate, login, logout

@strawberry.type
class LoginMutation:

    @strawberry.mutation
    def login(info, username: str, password: str) -> str:
        request = info.context['request']
        user = authenticate(request, username=username, password=password)
        if user is not None:
            login(request, user)
            return 'ok'
        logout(request)
        return 'failed'

    @strawberry.mutation
    def logout(info) -> str:
        request = info.context['request']
        logout(request)
        return 'ok'




from decimal import Decimal
from django.db.models import fields, Model
from typing import List, Optional, cast
from uuid import UUID
import datetime
import strawberry
from . import utils

field_type_map = {
    fields.AutoField: strawberry.ID,
    fields.BigAutoField: strawberry.ID,
    fields.BigIntegerField: int,
    fields.BooleanField: bool,
    fields.CharField: str,
    fields.DateField: datetime.date,
    fields.DateTimeField: datetime.datetime,
    fields.DecimalField: Decimal,
    fields.EmailField: str,
    #TODO: fields.FieldFile
    fields.FilePathField: str,
    fields.FloatField: float,
    #TODO: fields.ImageField
    fields.GenericIPAddressField: str,
    fields.IntegerField: int,
    #TODO: fields.JSONField
    fields.NullBooleanField: Optional[bool],
    fields.PositiveBigIntegerField: int,
    fields.PositiveIntegerField: int,
    fields.PositiveSmallIntegerField: int,
    fields.SlugField: str,
    fields.SmallAutoField: strawberry.ID,
    fields.SmallIntegerField: int,
    fields.TimeField: datetime.time,
    fields.URLField: str,
    fields.UUIDField: UUID,
    fields.TextField: str,
}

def get_field_type(field):
    db_field_type = type(field)
    field_type = field_type_map.get(db_field_type)
    if field_type is None:
        raise TypeError(f'Unknown type for {db_field_type.__name__}')
    return field_type


model_type_map = {}

def register_model_type(model, field_type):
    #if model in model_type_map:
    #    raise AttributeError(f'Model {model} already registered')
    model_type_map[model] = field_type

def get_model_type(model):
    model_type = model_type_map.get(model)
    if model_type is None:
        name = model._meta.object_name
        package, module = model.__module__.split('.', 1)
        model_type = LazyModelType(name, module, package)
        model_type.model = model
    return model_type

class LazyModelType(strawberry.LazyType):
    def resolve_type(self):
        if self.model not in model_type_map:
            raise Exception(f'GraphQL type not defined for "{self.model._meta.object_name}" Django model.')
        return model_type_map[self.model]


resolver_cls_map = {}

def register_resolver_cls(model, resolver_cls):
    #if model in resolver_cls_map:
    #    raise AttributeError(f'Model {model} already registered')
    resolver_cls_map[model] = resolver_cls

def get_resolver_cls(model):
    resolver_cls = resolver_cls_map.get(model)
    if not resolver_cls:
        raise TypeError(f'Model resolver not defined for {model}')
    return resolver_cls


def get_field(field, is_input, is_update):
    field_type = get_field_type(field)

    if is_input and field_type == strawberry.ID:
        return #TODO: is this correct?

    optional = False
    if is_input:
        has_default = field.default != fields.NOT_PROVIDED
        if field.blank or is_update or has_default:
            optional = True
    if field.null:
        optional = True

    if optional:
        field_type = Optional[field_type]

    return field.name, field_type, {}


def get_relation_field(field):
    if hasattr(field, 'get_accessor_name'):
        field_name = field.get_accessor_name()
    else:
        field_name = field.name
    model = field.related_model
    field_type = get_model_type(model)

    def resolver(info, root, filters: Optional[List[str]] = None) -> List[field_type]:
        #resolver_cls = get_resolver_cls(model)
        #instance = resolver_cls(info, root)
        qs = getattr(root, field_name).all()
        if filters:
            filters, excludes = utils.split_filters(filters)
            if filters:
                qs = qs.filter(**filters)
            if excludes:
                qs = qs.exclude(**excludes)
        return qs


    return field_name, None, {'resolver': resolver}


def get_relation_foreignkey_field(field, is_input):
    field_name = field.name
    model = field.related_model
    field_type = get_model_type(model)

    if is_input:
        field_type = strawberry.ID
        if field.blank or field.null:
            field_type = Optional[field_type]
        return field_name, field_type, {}

    def resolver(info, root) -> Optional[field_type]:
        obj = getattr(root, field_name)
        return obj

    return field_name, None, {'resolver': resolver}


def is_in(item, item_list):
    if not item_list:
        return False
    return item in item_list


def generate_model_type(resolver_cls, is_input=False, is_update=False):
    model = resolver_cls.model
    annotations = {}
    attributes = { '__annotations__': annotations }

    # add fields
    for field in model._meta.get_fields():
        if resolver_cls.fields and field.name not in resolver_cls.fields:
            continue # skip
        if is_in(field.name, resolver_cls.exclude):
            continue # skip
        if is_input and is_in(field.name, resolver_cls.readonly_fields):
            continue # skip
        if field.is_relation:
            if isinstance(field, fields.related.ForeignKey):
                field_params = get_relation_foreignkey_field(field, is_input)
            else:
                if is_input:
                    continue
                field_params = get_relation_field(field)
        else:
            field_params = get_field(field, is_input, is_update)
        if not field_params:
            continue

        field_name, field_type, field_kwargs = field_params

        if is_input:
            attributes[field_name] = strawberry.arguments.UNSET
        else:
            if resolver_cls.field_permission_classes:
                field_kwargs['permission_classes'] = resolver_cls.field_permission_classes
            attributes[field_name] = strawberry.field(**field_kwargs)

        if field_type:
            annotations[field_name] = field_type

    if not is_input:
        for field_name in dir(resolver_cls):
            field = getattr(resolver_cls, field_name)
            if hasattr(field, '_field_definition'):
                attributes[field_name] = field

    # generate type
    type_name = model._meta.object_name
    if is_update:
        type_name = f'Update{type_name}'
    elif is_input:
        type_name = f'Create{type_name}'
    model_type = type(type_name, (), attributes)
    model_type = strawberry.type(model_type, is_input=is_input)
    if not is_input:
        register_model_type(model, model_type)
        register_resolver_cls(model, resolver_cls)
    return model_type



import strawberry


class ModelPermissions(strawberry.BasePermission):
    def has_permission(self, root, info, **kwargs):
        if info.context is None:
            self.message = f'Missing context object'
            return False
        request = info.context.get('request')
        if request is None:
            self.message = f'Missing request object'
            return False
        # required permission hass been added to this object by wrap_model_permission_class
        if not request.user.has_perms([self.required_permission]):
            self.message = f'User does not have {self.required_permission} permission'
            return False
        return True


def permission_class_wrapper(permission_class, model, permission_codename):
    if not issubclass(permission_class, ModelPermissions):
        return permission_class

    app_label, model_name = model._meta.app_label, model._meta.model_name
    permission_name = f'{app_label}.{permission_codename}_{model_name}'
    def _permission_class(*args, **kwargs):
        instance = permission_class(*args, **kwargs)
        # this will be checked later on in ModelPermissions class
        instance.required_permission = permission_name
        return instance
    return _permission_class



from typing import List, Optional, cast
import strawberry
from . import utils
from .types import generate_model_type
from .permissions import permission_class_wrapper
from django.core import exceptions


class ModelResolverMixin:
    field_permission_classes = None
    permission_classes = None
    fields = None
    exclude = None
    readonly_fields = None

    def __init__(self, info, root):
        self.info = info
        self.root = root
        self.request = None
        if info.context:
            self.request = info.context.get('request')

    def get_model(self):
        return self.model

    def get_queryset(self):
        model = self.get_model()
        return model.objects.all()

    def get_queryset_filtered(self, filters):
        qs = self.get_queryset()
        if not filters:
            return qs
        filters, excludes = utils.split_filters(filters)
        if filters:
            qs = qs.filter(**filters)
        if excludes:
            qs = qs.exclude(**excludes)
        return qs

    def list(self, **filters):
        qs = self.get_queryset_filtered(**filters)
        return qs

    def get(self, id):
        qs = self.get_queryset()
        return qs.get(id=id)

    def create(self, data):
        model = self.get_model()
        instance = model(**data)
        instance.save()
        return instance

    def update(self, data, **filters):
        qs = self.get_queryset_filtered(**filters)
        qs.update(**data)
        return qs

    def delete(self, **filters):
        qs = self.get_queryset_filtered(**filters)
        items = list(qs)
        qs.delete()
        return items


def get_permission_classes(cls, permission_codename):
    if not cls.permission_classes:
        return None
    permission_classes = [
        permission_class_wrapper(permission_class, cls.model, permission_codename)
            for permission_class in cls.permission_classes
    ]
    return permission_classes


class ModelFieldMixin:
    @classmethod
    def get_field(cls):
        permission_classes = get_permission_classes(cls, 'view')

        @strawberry.field(permission_classes=permission_classes)
        def get_field(info, root, id: strawberry.ID) -> cls.output_type:
            instance = cls(info, root)
            return instance.get(id)

        return get_field

    @classmethod
    def list_field(cls):
        permission_classes = get_permission_classes(cls, 'view')

        @strawberry.field(permission_classes=permission_classes)
        def list_field(info, root,
                filters: Optional[List[str]] = None) -> List[cls.output_type]:
            instance = cls(info, root)
            return instance.list(filters=filters)

        return list_field

    @classmethod
    def query_fields(cls):
        return cls.get_field(), cls.list_field()


class ModelMutationMixin:
    @classmethod
    def create_mutation(cls):
        permission_classes = get_permission_classes(cls, 'add')

        @strawberry.mutation(permission_classes=permission_classes)
        def create_mutation(info, root, data: cls.create_input_type) -> cls.output_type:
            instance = cls(info, root)
            return instance.create(utils.get_data(cls.model, data))
        return create_mutation

    @classmethod
    def update_mutation(cls):
        permission_classes = get_permission_classes(cls, 'change')

        @strawberry.mutation(permission_classes=permission_classes)
        def update_mutation(info, root, data: cls.update_input_type,
                filters: Optional[List[str]] = None) -> List[cls.output_type]:
            instance = cls(info, root)
            return instance.update(utils.get_data(cls.model, data), filters=filters)

        return update_mutation

    @classmethod
    def delete_mutation(cls):
        permission_classes = get_permission_classes(cls, 'delete')

        @strawberry.mutation(permission_classes=permission_classes)
        def delete_mutation(info, root,
                filters: Optional[List[str]] = None) -> List[cls.output_type]:
            instance = cls(info, root)
            return instance.delete(filters=filters)

        return delete_mutation

    @classmethod
    def mutations(cls):
        return cls.create_mutation(), cls.update_mutation(), cls.delete_mutation()


class ModelQueryMixin:
    @classmethod
    def query(cls):
        object_name = utils.camel_to_snake(cls.model._meta.object_name)
        class Query: pass
        setattr(Query, f'{object_name}', cls.get_field())
        setattr(Query, f'{object_name}s', cls.list_field())
        return strawberry.type(Query)

    @classmethod
    def mutation(cls):
        object_name = utils.camel_to_snake(cls.model._meta.object_name)
        class Mutation: pass
        setattr(Mutation, f'create_{object_name}', cls.create_mutation())
        setattr(Mutation, f'update_{object_name}s', cls.update_mutation())
        setattr(Mutation, f'delete_{object_name}s', cls.delete_mutation())
        return strawberry.type(Mutation)


class ModelTypeBase(type):
    def __new__(cls, name, bases, attrs, **kwargs):
        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        if hasattr(new_cls, 'model'):
            new_cls.output_type = generate_model_type(new_cls)
            new_cls.create_input_type = generate_model_type(new_cls, is_input=True)
            new_cls.update_input_type = generate_model_type(new_cls, is_input=True, is_update=True)
        return new_cls


class ModelResolver(
        ModelResolverMixin,
        ModelFieldMixin,
        ModelMutationMixin,
        ModelQueryMixin,
        metaclass=ModelTypeBase):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['strawberry_django']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.1,<4.0', 'strawberry-graphql>=0.42']

setup_kwargs = {
    'name': 'strawberry-graphql-django',
    'version': '0.0.5',
    'description': 'Strawberry GraphQL Django extension',
    'long_description': '# Strawberry GraphQL Django extension\n\nThis library provides helpers to generate fields, mutations and resolvers for Django models.\n\nInstalling strawberry-graphql-django packet from the python package repository.\n```shell\npip install strawberry-graphql-django\n```\n\n## Example project files\n\nSee example Django project [examples/django](examples/django).\n\nmodels.py:\n```python\nfrom django.db import models\n\nclass User(models.Model):\n    name = models.CharField(max_length=50)\n    age = models.IntegerField()\n    groups = models.ManyToManyField(\'Group\', related_name=\'users\')\n\nclass Group(models.Model):\n    name = models.CharField(max_length=50)\n```\n\nschema.py:\n```python\nimport strawberry\nfrom strawberry_django import ModelResolver, ModelPermissions\nfrom .models import User, Group\n\nclass UserResolver(ModelResolver):\n    model = User\n    @strawberry.field\n    def age_in_months(info, root) -> int:\n        return root.age * 12\n\nclass GroupResolver(ModelResolver):\n    model = Group\n    fields = [\'name\', \'users\']\n\n    # only users who have group permissions can access and modify groups\n    permissions_classes = [ModelPermissions]\n\n    # queryset filtering\n    def get_queryset(self):\n        qs = super().get_queryset()\n        # only super users can access groups\n        if not self.request.user.is_superuser:\n            qs = qs.none()\n        return qs\n\n@strawberry.type\nclass Query(UserResolver.query(), GroupResolver.query()):\n    pass\n\n@strawberry.type\nclass Mutation(UserResolver.mutation(), GroupResolver.mutation()):\n    pass\n\nschema = strawberry.Schema(query=Query, mutation=Mutation)\n```\n\nurls.py:\n```python\nfrom django.urls import include, path\nfrom strawberry.django.views import GraphQLView\nfrom .schema import schema\n\nurlpatterns = [\n    path(\'graphql\', GraphQLView.as_view(schema=schema)),\n]\n```\n\nAdd models and schema. Create database. Start development server.\n```shell\nmanage.py makemigrations\nmanage.py migrate\nmanage.py runserver\n```\n\n## Mutations and Queries\n\nOpen http://localhost:8000/graphql and start testing.\n\nCreate new user.\n```\nmutation {\n  createUser(data: {name: "my user", age: 20}) {\n    id\n  }\n}\n```\n\nMake first queries.\n```\nquery {\n  user(id: 1) {\n    name\n    age\n    groups {\n        name\n    }\n  }\n  users(filters: ["name__contains=my", "!age__gt=60"]) {\n    id\n    name\n    ageInMonths\n  }\n}\n```\n\nUpdate user data.\n```\nmutation {\n  updateUsers(data: {name: "new name"}, filters: ["id=1"]) {\n    id\n    name\n  }\n}\n```\n\nFinally delete user.\n```\nmutation {\n  deleteUsers(filters: ["id=1"]) {\n    id\n  }\n}\n```\n\n## Contributing\n\nI would be more than happy to get pull requests, improvement ideas or any feedback from you.\n',
    'author': 'Lauri Hintsala',
    'author_email': 'lauri.hintsala@verkkopaja.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/la4de/strawberry-graphql-django',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import strawberry
from django.db.models import fields

def split_filters(filters):
    filter, exclude = {}, {}
    for string in filters:
        try:
            k, v = string.split('=', 1)
        except ValueError:
            raise ValueError(f'Invalid filter "{filter}"')
        if k.startswith('!'):
            k = k.strip('!')
            exclude[k] = v
        else:
            filter[k] = v
    return filter, exclude

def get_data(model, data):
    values = {}
    for field in model._meta.fields:
        field_name = field.name
        value = getattr(data, field.name, strawberry.arguments.UNSET)
        if value is strawberry.arguments.UNSET:
            continue
        if isinstance(field, fields.related.ForeignKey):
            field_name = f'{field_name}_id'
        values[field_name] = value
    return values

def camel_to_snake(s):
    return ''.join(['_'+c.lower() if c.isupper() else c for c in s]).lstrip('_')
